#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/fcntl.h>
#include <sys/bus.h>
#include <sys/module.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/tty.h>
#include <sys/selinfo.h>
#include <sys/proc.h>
#include <sys/sysctl.h>
#include <sys/uio.h>

#include <dev/usb/usb.h>
#include <dev/usb/usbhid.h>

#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#include <dev/usb/hid.h>

#ifdef USB_DEBUG
#define DPRINTF(x)	if (uhjdebug) printf x
#define DPRINTFN(n,x)	if (uhjdebug>(n)) printf x
int	uhjdebug = 0;
SYSCTL_NODE(_hw_usb, OID_AUTO, uhj, CTLFLAG_RW, 0, "USB uhj");
SYSCTL_INT(_hw_usb_uhj, OID_AUTO, debug, CTLFLAG_RW,
	   &uhjdebug, 0, "uhj debug level");
#else
#define DPRINTF(x)
#define DPRINTFN(n,x)
#endif

#define UHJ_SDEVNUM 2

struct uhj_softc {
	device_t sc_dev;		/* base device */
	usbd_device_handle sc_udev;
	usbd_interface_handle sc_iface;	/* interface */
	usbd_pipe_handle sc_intrpipe;	/* interrupt pipe */
	int sc_ep_addr;

	int sc_isize;
	int sc_osize;
	int sc_fsize;

	u_char *sc_ibuf;
	u_char *sc_obuf;

	void *sc_repdesc;
	int sc_repdesc_size;

	int sc_ocnt;		/* number of opened "sub device" */
	u_char sc_open[UHJ_SDEVNUM]; /* flag indicate "sub device" is open */
	struct clist sc_q[UHJ_SDEVNUM];
	struct selinfo sc_rsel[UHJ_SDEVNUM];
	struct proc *sc_async[UHJ_SDEVNUM];	/* process that wants SIGIO */
	u_char sc_pstate;		/* "parent device" state */
#define UHJ_NEEDCLEAR	0x04	/* needs clearing endpoint stall */
	u_char sc_state[UHJ_SDEVNUM];	/* "sub device" state */
#define	UHJ_OPEN	0x01	/* device is open */
#define	UHJ_ASLP	0x02	/* waiting for device data */
#define UHJ_IMMED	0x08	/* return read data immediately */

	int sc_refcnt[UHJ_SDEVNUM];
	u_char sc_dying;

	struct cdev *dev[UHJ_SDEVNUM];
};

#define	UHJUNIT(dev)	(minor(dev))
#define	UHJ_CHUNK	128	/* chunk size for read */
#define	UHJ_BSIZE	1020	/* buffer size */

static const struct uhj_product {
	uint16_t vendor;
	uint16_t product;
} uhj_products[] = {
	{0x0e8f, 0x3013},
	{0x04d9, 0x0002},
	{0, 0}
};

d_open_t	uhjopen;
d_close_t	uhjclose;
d_read_t	uhjread;
d_write_t	uhjwrite;
d_ioctl_t	uhjioctl;
d_poll_t	uhjpoll;


static struct cdevsw uhj_cdevsw __unused = {
	.d_version =	D_VERSION,
	.d_flags =	D_NEEDGIANT,
	.d_open =	uhjopen,
	.d_close =	uhjclose,
	.d_read =	uhjread,
	.d_write =	uhjwrite,
	.d_ioctl =	uhjioctl,
	.d_poll =	uhjpoll,
	.d_name =	"uhj",
};

static void uhid_intr(usbd_xfer_handle xfer, usbd_private_handle addr,
    usbd_status status);
static int uhj_report_size(void *buf, int len, enum hid_kind k, u_int8_t id);

MODULE_DEPEND(uhj, usb, 1, 1, 1);

static device_probe_t uhj_match;
static device_attach_t uhj_attach;
static device_detach_t uhj_detach;

static device_method_t uhj_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,	uhj_match),
	DEVMETHOD(device_attach, uhj_attach),
	DEVMETHOD(device_detach, uhj_detach),

	{ 0, 0 }
};

static driver_t uhj_driver = {
	"uhj",
	uhj_methods,
	sizeof(struct uhj_softc)
};

static devclass_t uhj_devclass;

DRIVER_MODULE(uhj, uhub, uhj_driver, uhj_devclass, usbd_driver_load, 0);

static int
uhj_match(device_t self)
{
	struct usb_attach_arg *uaa = device_get_ivars(self);
	int i;

	/*
	 * XXX there are two device to match, one has iface,
	 * the other does not.
	 */
	if (uaa->iface == NULL)
		return(UMATCH_NONE);

	for (i = 0; uhj_products[i].vendor != 0; i++)
		if (uaa->vendor == uhj_products[i].vendor &&
		    uaa->product == uhj_products[i].product)
			return(UMATCH_VENDOR_PRODUCT);

	return(UMATCH_NONE);
}

static int
uhj_attach(device_t self)
{
	struct uhj_softc *sc = device_get_softc(self);
	struct usb_attach_arg *uaa = device_get_ivars(self);
	usbd_interface_handle iface = uaa->iface;
	usb_endpoint_descriptor_t *ed;
	void *desc;
	usbd_status err;
	int i, size, isize;

	sc->sc_dev = self;
	sc->sc_udev = uaa->device;
	sc->sc_iface = iface;

	ed = usbd_interface2endpoint_descriptor(iface, 0);
	if (ed == NULL) {
		printf("%s: could not read endpoint descriptor\n",
		    device_get_nameunit(sc->sc_dev));
		sc->sc_dying = 1;
		return ENXIO;
	}

	DPRINTFN(10,("uhj_attach: bLength=%d bDescriptorType=%d "
		     "bEndpointAddress=%d-%s bmAttributes=%d wMaxPacketSize=%d"
		     " bInterval=%d\n",
		     ed->bLength, ed->bDescriptorType,
		     ed->bEndpointAddress & UE_ADDR,
		     UE_GET_DIR(ed->bEndpointAddress)==UE_DIR_IN? "in" : "out",
		     ed->bmAttributes & UE_XFERTYPE,
		     UGETW(ed->wMaxPacketSize), ed->bInterval));

	if (UE_GET_DIR(ed->bEndpointAddress) != UE_DIR_IN ||
	    (ed->bmAttributes & UE_XFERTYPE) != UE_INTERRUPT) {
		printf("%s: unexpected endpoint\n",
		    device_get_nameunit(sc->sc_dev));
		sc->sc_dying = 1;
		return ENXIO;
	}

	sc->sc_ep_addr = ed->bEndpointAddress;

	desc = NULL;
	err = usbd_read_report_desc(uaa->iface, &desc, &size, M_USBDEV);
	if (err) {
		printf("%s: no report descriptor\n",
		    device_get_nameunit(sc->sc_dev));
		sc->sc_dying = 1;
		return ENXIO;
	}
	DPRINTFN(10,("uhj_attach: rdesc_size=%d\n", size));

	(void)usbd_set_idle(iface, 0, 0);

	/* report of the two "sub-device" should be identical except id */
	sc->sc_isize = uhj_report_size(desc, size, hid_input, 1) + 1;
	sc->sc_osize = uhj_report_size(desc, size, hid_output, 1) + 1;
/* 	sc->sc_fsize = uhj_report_size(desc, size, hid_feature, 1) + 1; */

	DPRINTFN(10,("uhj_attach: sc_isize=%d sc_osize=%d\n",
	    sc->sc_isize, sc->sc_osize));

	/* identical check */
	isize = uhj_report_size(desc, size, hid_input, 2);
	if (isize + 1 != sc->sc_isize) {
		printf("%s: report descriptor of sub-device not identical\n",
		    device_get_nameunit(sc->sc_dev));
		return ENXIO;
	}

	sc->sc_repdesc = desc;
	sc->sc_repdesc_size = size;

	/* FIXME hard-coded device unit number. */
	for (i = 0; i < UHJ_SDEVNUM; i++)
		sc->dev[i] = make_dev(&uhj_cdevsw, i, UID_ROOT, GID_OPERATOR,
		    0644, "uhid%d", i);

	return(0);
}

static int
uhj_detach(device_t self)
{
	struct uhj_softc *sc = device_get_softc(self);
	int i, s;

	DPRINTF(("uhj_detach: sc=%p\n", sc));
	sc->sc_dying = 1;
	if (sc->sc_intrpipe != NULL)
		usbd_abort_pipe(sc->sc_intrpipe);

	for (i = 0; i < UHJ_SDEVNUM; i++) {
		if (sc->sc_state[i] & UHJ_OPEN) {
			s = splusb();
			if (--sc->sc_refcnt[i] >= 0) {
				/* Wake everyone */
				wakeup(&sc->sc_q[i]);
				/* Wait for processes to go away. */
				usb_detach_wait(sc->sc_dev);
			}
			splx(s);
		}
		destroy_dev(sc->dev[i]);
	}

	if (sc->sc_repdesc)
		free(sc->sc_repdesc, M_USBDEV);

	return(0);
}

static int
uhj_report_size(void *buf, int len, enum hid_kind k, u_int8_t id)
{
	struct hid_data *d;
	struct hid_item h;
	int hi, lo, size;

	hi = lo = -1;
	for (d = hid_start_parse(buf, len, 1<<k); hid_get_item(d, &h); )
		if (h.kind == k && h.report_ID == id) {
			if (lo < 0)
				lo = h.loc.pos;
			hi = h.loc.pos + h.loc.size * h.loc.count;
		}
	hid_end_parse(d);
	size = hi - lo;

	return((size + 7) / 8);
}

static void
uhid_intr(usbd_xfer_handle xfer, usbd_private_handle addr, usbd_status status)
{
	struct uhj_softc *sc = addr;
	int i;


#ifdef USB_DEBUG
	if (uhjdebug > 5) {
		u_int32_t cc, j;

		usbd_get_xfer_status(xfer, NULL, NULL, &cc, NULL);
		DPRINTF(("uhj_intr: status=%d cc=%d\n", status, cc));
		DPRINTF(("uhj_intr: data ="));
		for (j = 0; j < cc; j++)
			DPRINTF((" %02x", sc->sc_ibuf[j]));
		DPRINTF(("\n"));
	}
#endif

	if (status == USBD_CANCELLED)
		return;

	if (status != USBD_NORMAL_COMPLETION) {
		DPRINTF(("uhj_intr: status=%d\n", status));
		if (status == USBD_STALLED)
		    sc->sc_pstate |= UHJ_NEEDCLEAR;
		return;
	}

	/* Get report ID from first byte. */
	i = sc->sc_ibuf[0] - 1;	/* Hard-coded FIXME */

	/* Discard it if corresponding "sub device" not open yet. */
	if (!sc->sc_open[i]) {
		DPRINTF(("uhj_intr: discard data for closed device\n"));
		return;
	}

	/*
	 * Trim report id field from data, make it more similar to
	 * a "real" no-report-id uhid(4) device.
	 */
	(void) b_to_q(&sc->sc_ibuf[1], sc->sc_isize - 1, &sc->sc_q[i]);

	if (sc->sc_state[i] & UHJ_ASLP) {
		sc->sc_state[i] &= ~UHJ_ASLP;
		DPRINTFN(5, ("uhj_intr: waking %p\n", &sc->sc_q[i]));
		wakeup(&sc->sc_q[i]);
	}
	selwakeuppri(&sc->sc_rsel[i], PZERO);
	if (sc->sc_async[i] != NULL) {
		DPRINTFN(3, ("uhj_intr: sending SIGIO %p\n", sc->sc_async[i]));
		PROC_LOCK(sc->sc_async[i]);
		psignal(sc->sc_async[i], SIGIO);
		PROC_UNLOCK(sc->sc_async[i]);
	}
}

int
uhjopen(struct cdev *dev, int flag, int mode, struct thread *p)
{
	struct uhj_softc *sc;
	usbd_status err;
	dev_t m;

	/* FIXME can't assume that there is always only one uhj device. */
	sc = devclass_get_softc(uhj_devclass, 0);
	if (sc == NULL)
		return (ENXIO);

	m = UHJUNIT(dev);
	DPRINTF(("uhjopen: sc=%p devname=%s minor=%d\n", sc, dev->si_name, m));

	if (sc->sc_dying)
		return (ENXIO);

	/* FIXME m is not always 0 or 1. */
	if (sc->sc_state[m] & UHJ_OPEN)
		return (EBUSY);
	sc->sc_state[m] |= UHJ_OPEN;

	clist_alloc_cblocks(&sc->sc_q[m], UHJ_BSIZE, UHJ_BSIZE); /* FIXME */
	if (sc->sc_ocnt == 0) {
		sc->sc_ibuf = malloc(sc->sc_isize, M_USBDEV, M_WAITOK);
		sc->sc_obuf = malloc(sc->sc_osize, M_USBDEV, M_WAITOK);
	}

	sc->sc_open[m] = 1;	/* FIXME */
	sc->sc_ocnt++;

	/*
	 * Two "uhid" sub devices share one interupt pipe, so we
	 * actually only need to set up intrpipe once.
	 */
	if (sc->sc_intrpipe != NULL) {
		DPRINTFN(10, ("uhjopen: fake open on %s\n", dev->si_name));
		return(0);
	}

	/* Set up interrupt pipe. */
	err = usbd_open_pipe_intr(sc->sc_iface, sc->sc_ep_addr,
	    USBD_SHORT_XFER_OK, &sc->sc_intrpipe, sc, sc->sc_ibuf,
	    sc->sc_isize, uhid_intr, USBD_DEFAULT_INTERVAL);
	if (err) {
		DPRINTF(("uhidopen: usbd_open_pipe_intr failed, "
			 "error=%d\n",err));
		free(sc->sc_ibuf, M_USBDEV);
		free(sc->sc_obuf, M_USBDEV);
		sc->sc_ibuf = sc->sc_obuf = NULL;
		sc->sc_state[m] &= ~UHJ_OPEN; /* FIXME */
		sc->sc_open[m] = 0;
		sc->sc_ocnt--;
		return (EIO);
	}

	sc->sc_state[m] &= ~UHJ_IMMED; /* FIXME */

	sc->sc_async[m] = 0;

	return(0);
}

int
uhjclose(struct cdev *dev, int flag, int mode, struct thread *p)
{
	struct uhj_softc *sc;
	dev_t m;

	/* FIXME */
	sc = devclass_get_softc(uhj_devclass, 0);

	m = UHJUNIT(dev);
	DPRINTF(("uhidclose: sc=%p devname=%s\n", sc, dev->si_name));
	sc->sc_open[m] = 0;	/* FIXME */
	sc->sc_ocnt--;

	/* Close interrupt pipe if all "sub devices" are closed. */
	if (sc->sc_ocnt == 0) {
		usbd_abort_pipe(sc->sc_intrpipe);
		usbd_close_pipe(sc->sc_intrpipe);
		sc->sc_intrpipe = NULL;
	}

	ndflush(&sc->sc_q[m], sc->sc_q[m].c_cc); /* FIXME */
	clist_free_cblocks(&sc->sc_q[m]);

	if (sc->sc_ocnt == 0) {
		free(sc->sc_ibuf, M_USBDEV);
		free(sc->sc_obuf, M_USBDEV);
		sc->sc_ibuf = sc->sc_obuf = NULL;
	}

	sc->sc_state[m] &= ~UHJ_OPEN; /* FIXME */

	sc->sc_async[m] = 0;

	return(0);
}

int
uhjread(struct cdev *dev, struct uio *uio, int flag)
{
	struct uhj_softc *sc;
	dev_t m;
	size_t length;
	u_char buffer[UHJ_CHUNK];
	int s, error = 0;

	/* FIXME */
	sc = devclass_get_softc(uhj_devclass, 0);
	m = UHJUNIT(dev);

/* 	sc->sc_refcnt[m]++;	/\* FIXME *\/ */

	DPRINTFN(1, ("uhjread\n"));

	/* Add UHJ_IMMED support here */

	s = splusb();
	while (sc->sc_q[m].c_cc == 0) { /* FIXME */
		if (flag & O_NONBLOCK) {
			splx(s);
			return (EWOULDBLOCK);
		}
		sc->sc_state[m] |= UHJ_ASLP; /* FIXME */
		DPRINTFN(5, ("uhjread: sleep on %p\n", &sc->sc_q));
		error = tsleep(&sc->sc_q[m], PZERO | PCATCH, "uhjrd", 0);
		DPRINTFN(5, ("uhjread: woke, error=%d\n", error));
		if (sc->sc_dying)
			error = EIO;
		if (error) {
			sc->sc_state[m] &= ~UHJ_ASLP; /* FIXME */
			break;
		}
		if (sc->sc_pstate & UHJ_NEEDCLEAR) {
			DPRINTFN(-1,("uhjread: clearing stall\n"));
			sc->sc_pstate &= ~UHJ_NEEDCLEAR;
			usbd_clear_endpoint_stall(sc->sc_intrpipe);
		}
	}
	splx(s);

	/* Transfer as many bytes as possible. */
	while (sc->sc_q[m].c_cc > 0 && uio->uio_resid > 0 && !error) { /* FIXME */
		length = min(sc->sc_q[m].c_cc, uio->uio_resid);
		if (length > sizeof(buffer))
			length = sizeof(buffer);

		/* Remove a small chunk from the input queue. */
		(void) q_to_b(&sc->sc_q[m], buffer, length);
		DPRINTFN(5, ("uhjread: got %lu chars\n", (u_long)length));

		/* Copy the data to the user process. */
		if ((error = uiomove(buffer, length, uio)) != 0)
			break;
	}

/* 	if (--sc->sc_refcnt[m] < 0) /\* FIXME *\/ */
/* 		usb_detach_wakeup(sc->sc_dev); */

	return(error);
}

int
uhjwrite(struct cdev *dev, struct uio *uio, int flag)
{
	/* not supported yet */
	return(0);
}

int
uhjioctl(struct cdev *dev, u_long cmd, caddr_t addr, int flag,
    struct thread *p)
{
	struct uhj_softc *sc;
	struct usb_ctl_report_desc *rd;
	const u_char *r;
	dev_t m;
	int size;

	/* FIXME */
	sc = devclass_get_softc(uhj_devclass, 0);
	m = UHJUNIT(dev);

/* 	sc->sc_refcnt[m]++;	/\* FIXME *\/ */

	DPRINTFN(2, ("uhidioctl: cmd=%lx\n", cmd));

	if (sc->sc_dying)
		return (EIO);

	switch(cmd) {
	case USB_GET_REPORT_DESC:
		rd = (struct usb_ctl_report_desc *)addr;
		r = sc->sc_repdesc;
		size = min(sc->sc_repdesc_size / UHJ_SDEVNUM,
		    sizeof(rd->ucrd_data));
		rd->ucrd_size = size;
		/* FIXME */
		memcpy(rd->ucrd_data, &r[size * m], size);
		break;
	case USB_GET_REPORT_ID:
		*(int *)addr = m + 1; /* FIXME */
		break;
	default:
		return (EINVAL);
	}

/* 	if (--sc->sc_refcnt[m] < 0) /\* FIXME *\/ */
/* 		usb_detach_wakeup(sc->sc_dev); */
	return(0);
}

int
uhjpoll(struct cdev *dev, int events, struct thread *p)
{
	/* not supported yet */
	return(0);
}
